# Copyright 2013,2015 Christoph Reiter, Dino Miniutti
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

"""Helpers for atomic file operations on Linux/OSX/Windows"""

import os
import contextlib

from senf import fsnative
from pathlib import Path

if os.name == "nt":
    from . import winapi
else:
    import fcntl

from .misc import NamedTemporaryFile


def _windows_rename(source, dest):
    """Replaces dest with source.

    Raises OSError in case of an error.
    """

    assert os.name == "nt"

    # not atomic, but better than removing the original first...
    status = winapi.MoveFileExW(
        source, dest, winapi.MOVEFILE_WRITE_THROUGH | winapi.MOVEFILE_REPLACE_EXISTING
    )

    if status == 0:
        raise winapi.WinError()


@contextlib.contextmanager
def atomic_save(filename, mode):
    """Try to replace the content of a file in the safest way possible.

    A temporary file will be created in the same directory where the
    replacement data can be written into.
    After writing is done the data will be flushed to disk and the original
    file replaced atomically.

    In case of an error this raises IOError and OSError and the original file
    will be untouched. In case the computer crashes or any other
    non-recoverable error happens the temporary file will be left behind and
    has to be deleted manually.

    with atomic_save("config.cfg", "wb") as f:
        f.write(data)
    """

    assert isinstance(filename, fsnative)

    dir_ = os.path.dirname(filename)
    basename = os.path.basename(filename)
    fileobj = NamedTemporaryFile(
        mode=mode,
        dir=dir_,
        prefix=basename + fsnative("_"),
        suffix=fsnative(".tmp"),
        delete=False,
    )

    try:
        yield fileobj

        fileobj.flush()
        fileno = fileobj.fileno()

        if os.name != "nt" and hasattr(fcntl, "F_FULLFSYNC"):
            # on OSX fsync doesn't sync all the way..
            # https://lists.apple.com/archives/darwin-dev/2005/Feb/msg00072.html
            # https://developer.apple.com/library/mac/documentation/Darwin/Reference/ManPages/man2/fsync.2.html
            fcntl.fcntl(fileno, fcntl.F_FULLFSYNC)
        else:
            # on linux fsync goes all the way by default
            # http://linux.die.net/man/2/fsync
            os.fsync(fileno)

        fileobj.close()

        # dereference symbolic links
        target = filename if not os.path.islink(filename) else Path(filename).resolve()

        if os.name == "nt":
            _windows_rename(fileobj.name, str(target))
        else:
            os.rename(fileobj.name, target)
    except Exception:
        try:
            os.unlink(fileobj.name)
        except OSError:
            pass
        raise
    finally:
        fileobj.close()



# Copyright 2014 Christoph Reiter
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

import unicodedata
import sys
from urllib.request import urlopen

from quodlibet.util import cached_func


_DIACRITIC_CACHE = {
    "\u0300": (
        "AEINOUWYaeinouwy\u0391\u0395\u0397\u0399\u039f\u03a5\u03a9"
        "\u03b1\u03b5\u03b7\u03b9\u03bf\u03c5\u03c9\u0415\u0418"
        "\u0435\u0438"
    ),
    "\u0300\u0345": "\u03b1\u03b7\u03c9",
    "\u0301": (
        "ACEGIKLMNOPRSUWYZacegiklmnoprsuwyz\xc6\xd8\xe6\xf8\u0391"
        "\u0395\u0397\u0399\u039f\u03a5\u03a9\u03b1\u03b5\u03b7"
        "\u03b9\u03bf\u03c5\u03c9\u0413\u041a\u0433\u043a"
    ),
    "\u0301\u0307": "Ss",
    "\u0301\u0345": "\u03b1\u03b7\u03c9",
    "\u0302": "ACEGHIJOSUWYZaceghijosuwyz",
    "\u0302\u0300": "AEOaeo",
    "\u0302\u0301": "AEOaeo",
    "\u0302\u0303": "AEOaeo",
    "\u0302\u0309": "AEOaeo",
    "\u0303": "AEINOUVYaeinouvy",
    "\u0303\u0301": "OUou",
    "\u0303\u0304": "Oo",
    "\u0303\u0308": "Oo",
    "\u0304": (
        "AEGIOUYaegiouy\xc6\xe6\u0391\u0399\u03a5\u03b1\u03b9"
        "\u03c5\u0418\u0423\u0438\u0443"
    ),
    "\u0304\u0300": "EOeo",
    "\u0304\u0301": "EOeo",
    "\u0304\u0308": "Uu",
    "\u0306": (
        "AEGIOUaegiou\u0391\u0399\u03a5\u03b1\u03b9\u03c5\u0410"
        "\u0415\u0416\u0418\u0423\u0430\u0435\u0436\u0438\u0443"
    ),
    "\u0306\u0300": "Aa",
    "\u0306\u0301": "Aa",
    "\u0306\u0303": "Aa",
    "\u0306\u0309": "Aa",
    "\u0307": "ABCDEFGHIMNOPRSTWXYZabcdefghmnoprstwxyz",
    "\u0307\u0304": "AOao",
    "\u0308": (
        "AEHIOUWXYaehiotuwxy\u0399\u03a5\u03b9\u03c5\u0406\u0410"
        "\u0415\u0416\u0417\u0418\u041e\u0423\u0427\u042b\u042d"
        "\u0430\u0435\u0436\u0437\u0438\u043e\u0443\u0447\u044b"
        "\u044d\u0456\u04d8\u04d9\u04e8\u04e9"
    ),
    "\u0308\u0300": "Uu\u03b9\u03c5",
    "\u0308\u0301": "IUiu\u03b9\u03c5",
    "\u0308\u0304": "AOUaou",
    "\u0308\u030c": "Uu",
    "\u0308\u0342": "\u03b9\u03c5",
    "\u0309": "AEIOUYaeiouy",
    "\u030a": "AUauwy",
    "\u030a\u0301": "Aa",
    "\u030b": "OUou\u0423\u0443",
    "\u030c": "ACDEGHIKLNORSTUZacdeghijklnorstuz\u01b7\u0292",
    "\u030c\u0307": "Ss",
    "\u030f": "AEIORUaeioru\u0474\u0475",
    "\u0311": "AEIORUaeioru",
    "\u0313": (
        "\u0391\u0395\u0397\u0399\u039f\u03a9\u03b1\u03b5\u03b7"
        "\u03b9\u03bf\u03c1\u03c5\u03c9"
    ),
    "\u0313\u0300": (
        "\u0391\u0395\u0397\u0399\u039f\u03a9\u03b1\u03b5"
        "\u03b7\u03b9\u03bf\u03c5\u03c9"
    ),
    "\u0313\u0300\u0345": "\u0391\u0397\u03a9\u03b1\u03b7\u03c9",
    "\u0313\u0301": (
        "\u0391\u0395\u0397\u0399\u039f\u03a9\u03b1\u03b5"
        "\u03b7\u03b9\u03bf\u03c5\u03c9"
    ),
    "\u0313\u0301\u0345": "\u0391\u0397\u03a9\u03b1\u03b7\u03c9",
    "\u0313\u0342": "\u0391\u0397\u0399\u03a9\u03b1\u03b7\u03b9\u03c5\u03c9",
    "\u0313\u0342\u0345": "\u0391\u0397\u03a9\u03b1\u03b7\u03c9",
    "\u0313\u0345": "\u0391\u0397\u03a9\u03b1\u03b7\u03c9",
    "\u0314": (
        "\u0391\u0395\u0397\u0399\u039f\u03a1\u03a5\u03a9\u03b1"
        "\u03b5\u03b7\u03b9\u03bf\u03c1\u03c5\u03c9"
    ),
    "\u0314\u0300": (
        "\u0391\u0395\u0397\u0399\u039f\u03a5\u03a9\u03b1"
        "\u03b5\u03b7\u03b9\u03bf\u03c5\u03c9"
    ),
    "\u0314\u0300\u0345": "\u0391\u0397\u03a9\u03b1\u03b7\u03c9",
    "\u0314\u0301": (
        "\u0391\u0395\u0397\u0399\u039f\u03a5\u03a9\u03b1"
        "\u03b5\u03b7\u03b9\u03bf\u03c5\u03c9"
    ),
    "\u0314\u0301\u0345": "\u0391\u0397\u03a9\u03b1\u03b7\u03c9",
    "\u0314\u0342": ("\u0391\u0397\u0399\u03a5\u03a9\u03b1\u03b7\u03b9\u03c5\u03c9"),
    "\u0314\u0342\u0345": "\u0391\u0397\u03a9\u03b1\u03b7\u03c9",
    "\u0314\u0345": "\u0391\u0397\u03a9\u03b1\u03b7\u03c9",
    "\u031b": "OUou",
    "\u031b\u0300": "OUou",
    "\u031b\u0301": "OUou",
    "\u031b\u0303": "OUou",
    "\u031b\u0309": "OUou",
    "\u031b\u0323": "OUou",
    "\u0323": "ABDEHIKLMNORSTUVWYZabdehiklmnorstuvwyz",
    "\u0323\u0302": "AEOaeo",
    "\u0323\u0304": "LRlr",
    "\u0323\u0306": "Aa",
    "\u0323\u0307": "Ss",
    "\u0324": "Uu",
    "\u0325": "Aa",
    "\u0326": "STst",
    "\u0327": "CDEGHKLNRSTcdeghklnrst",
    "\u0327\u0301": "Cc",
    "\u0327\u0306": "Ee",
    "\u0328": "AEIOUaeiou",
    "\u0328\u0304": "Oo",
    "\u032d": "DELNTUdelntu",
    "\u032e": "Hh",
    "\u0330": "EIUeiu",
    "\u0331": "BDKLNRTZbdhklnrtz",
    "\u0342": "\u03b1\u03b7\u03b9\u03c5\u03c9",
    "\u0342\u0345": "\u03b1\u03b7\u03c9",
    "\u0345": "\u0391\u0397\u03a9\u03b1\u03b7\u03c9",
}

# See misc/uca_decomps.py
_UCA_DECOMPS_CACHE = {
    "AA": "\ua732",
    "AE": "\xc6\u01e2\u01fc",
    "AO": "\ua734",
    "AU": "\ua736",
    "AV": "\ua738\ua73a",
    "AY": "\ua73c",
    "D": "\xd0\u0110\ua779",
    "DZ": "\u01c4\u01f1",
    "Dz": "\u01c5\u01f2",
    "F": "\ua77b",
    "G": "\ua77d",
    "H": "\u0126",
    "IJ": "\u0132",
    "L": "\u0141",
    "LJ": "\u01c7",
    "LL": "\u1efa",
    "Lj": "\u01c8",
    "NJ": "\u01ca",
    "Nj": "\u01cb",
    "O": "\xd8\u01fe",
    "OE": "\u0152",
    "OO": "\ua74e",
    "R": "\ua782",
    "S": "\ua784",
    "SS": "\u1e9e",
    "T": "\ua786",
    "Tz": "\ua728",
    "VY": "\ua760",
    "aa": "\ua733",
    "ae": "\xe6\u01e3\u01fd",
    "ao": "\ua735",
    "au": "\ua737",
    "av": "\ua739\ua73b",
    "ay": "\ua73d",
    "d": "\xf0\u0111\ua77a",
    "db": "\u0238",
    "dz": "\u01c6\u01f3\u02a3",
    "d\u0291": "\u02a5",
    "d\u0292": "\u02a4",
    "f": "\ua77c",
    "ff": "\ufb00",
    "ffi": "\ufb03",
    "ffl": "\ufb04",
    "fi": "\ufb01",
    "fl": "\ufb02",
    "f\u014b": "\u02a9",
    "g": "\u1d79",
    "h": "\u0127\u210f",
    "ij": "\u0133",
    "l": "\u0142",
    "lj": "\u01c9",
    "ll": "\u1efb",
    "ls": "\u02aa",
    "lz": "\u02ab",
    "n": "\u0149",
    "nj": "\u01cc",
    "o": "\xf8\u01ff",
    "oe": "\u0153",
    "oo": "\ua74f",
    "qp": "\u0239",
    "r": "\ua783",
    "s": "\ua785",
    "ss": "\xdf",
    "st": "\ufb05\ufb06",
    "t": "\ua787",
    "th": "\u1d7a",
    "ts": "\u01be\u02a6",
    "tz": "\ua729",
    "t\u0255": "\u02a8",
    "t\u0283": "\u02a7",
    "vy": "\ua761",
    "zw": "\u018d",
    "\u039a\u03b1\u03b9": "\u03cf",
    "\u03ba\u03b1\u03b9": "\u03d7",
    "\u03c3": "\u03c2\u03f2\U0001d6d3\U0001d70d\U0001d747\U0001d781\U0001d7bb",
    "\u0413": "\u0490",
    "\u041e": "\ua668\ua66a\ua66c",
    "\u0433": "\u0491",
    "\u043e": "\ua669\ua66b\ua66d",
    "\u0565\u0582": "\u0587",
    "\u0574\u0565": "\ufb14",
    "\u0574\u056b": "\ufb15",
    "\u0574\u056d": "\ufb17",
    "\u0574\u0576": "\ufb13",
    "\u057e\u0576": "\ufb16",
    "\u2c95\u2c81\u2c93": "\u2ce4",
}


_PUNCT_CONFUSABLES_CACHE = {
    "!": "\uff01\u01c3\u2d51",
    "!!": "\u203c",
    "!?": "\u2049",
    '"': (
        "\u1cd3\uff02\u201c\u201d\u201f\u2033\u2036\u3003\u05f4\u02dd"
        "\u02ba\u02f6\u02ee\u05f2"
    ),
    "&": "\ua778",
    "'": (
        "\u055d\uff07\u2018\u2019\u201b\u2032\u2035\u055a\u05f3`\u1fef"
        "\uff40\xb4\u0384\u1ffd\u1fbd\u1fbf\u1ffe\u02b9\u0374\u02c8\u02ca"
        "\u02cb\u02f4\u02bb\u02bd\u02bc\u02be\ua78c\u05d9\u07f4\u07f5"
        "\u144a\u16cc"
    ),
    "''": (
        '\u1cd3"\uff02\u201c\u201d\u201f\u2033\u2036\u3003\u05f4\u02dd'
        "\u02ba\u02f6\u02ee\u05f2"
    ),
    "'''": "\u2034\u2037",
    "''''": "\u2057",
    "(": "\uff3b\u2768\u2772\u3014\ufd3e",
    "((": "\u2e28",
    ")": "\uff3d\u2769\u2773\u3015\ufd3f",
    "))": "\u2e29",
    "*": "\u204e\u066d\u2217\U0001031f",
    ",": "\u060d\u066b\u201a\xb8\ua4f9",
    "-": ("\u2010\u2011\u2012\u2013\ufe58\u06d4\u2043\u02d7\u2212\u2796\u2cba"),
    "-.": "\ua4fe",
    ".": "\U0001d16d\u2024\u0701\u0702\ua60e\U00010a50\u0660\u06f0\ua4f8",
    ".,": "\ua4fb",
    "..": "\u2025\ua4fa",
    "...": "\u2026",
    "/": (
        "\u1735\u2041\u2215\u2044\u2571\u27cb\u29f8\U0001d23a\u31d3\u3033"
        "\u2cc6\u30ce\u4e3f\u2f03"
    ),
    "//": "\u2afd",
    "///": "\u2afb",
    ":": (
        "\u0903\u0a83\uff1a\u0589\u0703\u0704\u16ec\ufe30\u1803\u1809"
        "\u205a\u05c3\u02f8\ua789\u2236\u02d0\ua4fd"
    ),
    ";": "\u037e",
    "?": "\u0294\u0241\u097d\u13ae\ua6eb",
    "?!": "\u2048",
    "??": "\u2047",
    "\\": (
        "\uff3c\ufe68\u2216\u27cd\u29f5\u29f9\U0001d20f\U0001d23b\u31d4\u4e36\u2f02"
    ),
    "\\\\": "\u2cf9\u244a",
    "_": "\u07fa\ufe4d\ufe4e\ufe4f",
    "{": "\u2774\U0001d114",
    "}": "\u2775",
}


def get_decomps_mapping(regenerate=False) -> dict[str, str]:
    """This takes the decomps.txt file of the Unicode UCA and gives us a cases
    where a letter can be decomposed for collation and that mapping isn't in
    NFKD.
    """

    if not regenerate:
        return _UCA_DECOMPS_CACHE

    mapping: dict[str, str] = {}

    h = urlopen("http://unicode.org/Public/UCA/8.0.0/decomps.txt")
    for line in h.read().splitlines():
        if line.startswith("#"):
            continue

        def to_uni(x):
            return chr(int(x, 16))

        def is_letter(x):
            return unicodedata.category(x) in ("Lu", "Ll", "Lt")

        cp, line = line.split(";", 1)
        tag, line = line.split(";", 1)
        decomp, line = line.split("#", 1)
        decomp = map(to_uni, decomp.strip().split())
        cp = to_uni(cp)

        if not is_letter(cp):
            continue

        decomp = filter(is_letter, decomp)
        simple = "".join(decomp)
        if not simple:
            continue

        # skip anything we get from normalization
        if unicodedata.normalize("NFKD", cp)[0] == simple:
            continue

        mapping[simple] = mapping.get(simple, "") + cp

    return mapping


def get_punctuation_mapping(regenerate=False) -> dict[str, str]:
    """This takes the unicode confusables set and extracts punctuation
    which looks similar to one or more ASCII punctuation.

    e.g. ' --> ＇

    """

    if not regenerate:
        return _PUNCT_CONFUSABLES_CACHE

    h = urlopen("http://www.unicode.org/Public/security/9.0.0/confusables.txt")
    data = h.read()
    mapping: dict[str, str] = {}
    for line in data.decode("utf-8-sig").splitlines():
        line = line.strip()
        if not line:
            continue
        if line.startswith("#"):
            continue

        char, repls = line.split(";", 2)[:2]
        char = char.strip()
        repls = repls.split()

        def to_uni(x):
            return chr(int(x, 16))

        char = to_uni(char)
        repls = [to_uni(r) for r in repls]

        def is_ascii(char):
            try:
                char.encode("ascii")
            except UnicodeEncodeError:
                return False
            return True

        def is_punct(char):
            return unicodedata.category(char).startswith("P")

        if all(is_ascii(c) and is_punct(c) for c in repls) and char:
            repls_joined = "".join(repls)
            mapping[repls_joined] = mapping.get(repls_joined, "") + char

    # if any of the equal chars is also ascii + punct we can replace
    # it as well
    for uni in mapping.values():
        also_ascii = [c for c in uni if is_ascii(c) and is_punct(c)]
        for c in also_ascii:
            mapping[c] = uni.replace(c, "")

    return mapping


def diacritic_for_letters(regenerate=False) -> dict[str, str]:
    """Returns a mapping for combining diacritic mark to ascii characters
    for which they can be used to combine to a single unicode char.

    (actually not ascii, but unicode from the Lu/Ll/Lt categories,
    but mainly ascii)

    Since this is quite expensive to compute, the result is a cached version
    unless regenerate != True. regenerate = True is used for unittests
    to validate the cache.
    """

    if not regenerate:
        return _DIACRITIC_CACHE

    d: dict[str, set[str]] = {}
    for i in range(sys.maxunicode):
        u = chr(i)
        n = unicodedata.normalize("NFKD", u)
        if len(n) <= 1:
            continue
        if unicodedata.category(u) not in ("Lu", "Ll", "Lt"):
            continue
        if not all(map(unicodedata.combining, n[1:])):
            continue
        d.setdefault(n[1:], set()).add(n[0])

    d2: dict[str, str] = {}
    for k, v in d.items():
        d2[k] = "".join(sorted(v))

    return d2


def generate_re_mapping(_diacritic_for_letters: dict[str, str]) -> dict[str, str]:
    letter_to_variants: dict[str, list[str]] = {}

    # combine combining characters with the ascii chars
    for dia, letters in _diacritic_for_letters.items():
        for c in letters:
            unichar = unicodedata.normalize("NFKC", c + dia)
            letter_to_variants.setdefault(c, []).append(unichar)

    letter_to_variants_joined: dict[str, str] = {}
    # create strings to replace ascii with
    for k, v in letter_to_variants.items():
        letter_to_variants_joined[k] = "".join(sorted(v))

    return letter_to_variants_joined


@cached_func
def get_replacement_mapping() -> dict[str, list[str]]:
    """Returns a dict mapping a sequence of characters to another sequence
    of characters.

    If a key occurs in a text, it should also match any of the characters in
    in the value.
    """

    mapping: dict[str, list[str]] = {}

    # use _DIACRITIC_CACHE and create a lookup table
    for cp, repl in generate_re_mapping(
        diacritic_for_letters(regenerate=False)
    ).items():
        mapping.setdefault(cp, []).extend(repl)

    # add more from the UCA decomp dataset
    for cp, repl in get_decomps_mapping(regenerate=False).items():
        mapping.setdefault(cp, []).extend(repl)

    # and some punctuation
    for cp, repl in get_punctuation_mapping(regenerate=False).items():
        mapping.setdefault(cp, []).extend(repl)

    return mapping



# Copyright 2017 Didier Villevalois
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

from gi.repository import Gtk

from quodlibet import _
from quodlibet.pattern import FileFromPattern
from quodlibet.plugins import PluginConfig, ConfProp
from quodlibet.plugins.playlist import PlaylistPlugin
from quodlibet.qltk import Icons
from quodlibet.qltk.entry import UndoEntry
from quodlibet.qltk.notif import Task
from quodlibet.qltk.window import Dialog
from quodlibet.qltk.msg import ErrorMessage
from quodlibet.util import copool
from quodlibet.util.dprint import print_d

from shutil import copyfile


class ExportToFolderDialog(Dialog):
    """A dialog to collect export settings"""

    def __init__(self, parent, pattern):
        super().__init__(
            title=_("Export Playlist to Folder"),
            transient_for=parent,
            use_header_bar=True,
        )

        self.set_default_size(400, -1)
        self.set_resizable(True)
        self.set_border_width(6)
        self.vbox.set_spacing(6)

        self.add_button(_("_Cancel"), Gtk.ResponseType.CANCEL)
        self.add_button(_("_Export"), Gtk.ResponseType.OK)
        self.set_default_response(Gtk.ResponseType.OK)

        box = Gtk.VBox(spacing=6)

        destination_label = Gtk.Label(_("Destination folder:"))
        destination_label.set_line_wrap(True)
        destination_label.set_xalign(0.0)
        box.pack_start(destination_label, False, False, 0)

        frame = Gtk.Frame()
        self.directory_chooser = Gtk.FileChooserWidget(
            action=Gtk.FileChooserAction.SELECT_FOLDER
        )
        self.directory_chooser.set_select_multiple(False)
        self.directory_chooser.set_border_width(1)
        frame.add(self.directory_chooser)
        frame.set_shadow_type(Gtk.ShadowType.IN)
        frame.set_border_width(0)
        box.pack_start(frame, True, True, 0)

        pattern_label = Gtk.Label(_("Filename pattern:"))
        pattern_label.set_line_wrap(True)
        pattern_label.set_xalign(0.0)
        box.pack_start(pattern_label, False, False, 0)

        self.pattern_entry = UndoEntry()
        self.pattern_entry.set_text(pattern)
        box.pack_start(self.pattern_entry, False, False, 0)

        self.vbox.pack_start(box, True, True, 0)

        self.set_response_sensitive(Gtk.ResponseType.OK, False)

        def changed(*args):
            has_directory = self.directory_chooser.get_filename() is not None
            self.set_response_sensitive(Gtk.ResponseType.OK, has_directory)

            pattern_text = self.pattern_entry.get_text()
            has_pattern = bool(pattern_text)
            self.set_response_sensitive(Gtk.ResponseType.OK, has_pattern)

        self.directory_chooser.connect("selection-changed", changed)
        self.pattern_entry.connect("changed", changed)

        self.get_child().show_all()


class Config:
    _config = PluginConfig(__name__)

    DEFAULT_PATTERN = "<artist~title>"

    default_pattern = ConfProp(_config, "default_pattern", DEFAULT_PATTERN)


CONFIG = Config()


class ExportToFolder(PlaylistPlugin):
    PLUGIN_ID = "ExportToFolder"
    PLUGIN_NAME = _("Export Playlist to Folder")
    PLUGIN_DESC = _("Exports a playlist by copying files to a folder.")
    PLUGIN_ICON = Icons.FOLDER
    REQUIRES_ACTION = True

    def __copy_songs(self, task, songs, directory, pattern, parent=None):
        """Generator for copool to copy songs to the folder"""
        self.__cancel = False
        total = len(songs)
        print_d(
            "Copying %d song(s) to directory %s. "
            "This might take a while..." % (total, directory)
        )
        for i, song in enumerate(songs):
            if self.__cancel:
                print_d("Cancelled export to directory.")
                self.__cancel = False
                break
            # Actually do the copy
            try:
                self._copy_file(song, directory, i + 1, pattern)
            except OSError as e:
                print_d(f"Unable to copy file: {e}")
                ErrorMessage(
                    parent,
                    _("Unable to export playlist"),
                    _("Ensure you have write access to the destination."),
                ).run()
                break
            task.update(float(i) / total)
            yield True
        print_d("Finished export to directory.")
        task.finish()

    def __cancel_copy(self):
        """Tell the copool to stop copying songs"""
        self.__cancel = True

    def _copy_file(self, song, directory, index, pattern):
        filename = song["~filename"]
        print_d(f"Copying {filename}.")
        new_name = pattern.format(song)
        copyfile(filename, "%s/%04d - %s" % (directory, index, new_name))

    def plugin_playlist(self, playlist):
        pattern_text = CONFIG.default_pattern
        dialog = ExportToFolderDialog(self.plugin_window, pattern_text)
        if dialog.run() == Gtk.ResponseType.OK:
            directory = dialog.directory_chooser.get_filename()
            pattern = FileFromPattern(dialog.pattern_entry.get_text())

            task = Task(
                "Export", _("Export Playlist to Folder"), stop=self.__cancel_copy
            )
            copool.add(
                self.__copy_songs,
                task,
                playlist.songs,
                directory,
                pattern,
                self.plugin_window,
                funcid="export-playlist-folder",
            )

        dialog.destroy()

    @classmethod
    def PluginPreferences(cls, parent):
        def changed(entry):
            CONFIG.default_pattern = entry.get_text()

        vbox = Gtk.VBox(spacing=6)

        def create_pattern():
            hbox = Gtk.HBox(spacing=6)
            hbox.set_border_width(6)
            label = Gtk.Label(label=_("Default filename pattern:"))
            hbox.pack_start(label, False, True, 0)
            entry = UndoEntry()
            if CONFIG.default_pattern:
                entry.set_text(CONFIG.default_pattern)
            entry.connect("changed", changed)
            hbox.pack_start(entry, True, True, 0)
            return hbox

        vbox.pack_start(create_pattern(), True, True, 0)

        return vbox



# Copyright 2014 Christoph Reiter
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

"""Some helper function for loading and converting image data."""

import math

from gi.repository import GdkPixbuf, Gtk, Gdk, GLib
import cairo


def get_surface_for_pixbuf(
    widget: Gtk.Widget, pixbuf: GdkPixbuf.Pixbuf | None
) -> cairo.Surface | None:
    """:returns: a cairo surface, if possible"""
    if not pixbuf:
        return None
    scale_factor = widget.get_scale_factor()
    return Gdk.cairo_surface_create_from_pixbuf(
        pixbuf, scale_factor, widget.get_window()
    )


def get_surface_extents(surface):
    """Gives (x, y, width, height) for a surface, scale independent"""

    ctx = cairo.Context(surface)
    x1, y1, x2, y2 = ctx.clip_extents()
    x1 = int(math.floor(x1))
    y1 = int(math.floor(y1))
    x2 = int(math.ceil(x2))
    y2 = int(math.ceil(y2))
    x2 -= x1
    y2 -= y1

    return (x1, y1, x2, y2)


def get_border_radius(_widgets=None):
    """Returns the border radius commonly used in the current theme.
    If there are no rounded corners 0 will be returned.
    """
    _widgets = _widgets or []
    if not _widgets:
        b = Gtk.Button()
        b.show()
        e = Gtk.Entry()
        e.show()
        _widgets += [b, e]

    radii = []
    for widget in _widgets:
        style_context = widget.get_style_context()
        radii.append(
            style_context.get_property(
                Gtk.STYLE_PROPERTY_BORDER_RADIUS, style_context.get_state()
            )
        )
    radius = max(radii)

    # Doesn't work on the default Ubuntu theme.
    # Not sure why, so fix manually for now
    theme_name = Gtk.Settings.get_default().props.gtk_theme_name
    if theme_name in ("Ambiance", "Radiance"):
        radius = int(radius / 1.5)

    return radius


def add_border(pixbuf, color, width=1, radius=0):
    """Add a border to the pixbuf and round of the edges.
    color is a Gdk.RGBA
    The resulting pixbuf will be width * 2px higher and wider.

    Can not fail.
    """

    pi = math.pi

    def new_border_path(ctx, w, h, r, margin=0):
        d = r + margin
        ctx.new_path()
        ctx.arc(w - d, d, r, -pi / 2, 0)
        ctx.arc(w - d, h - d, r, 0, pi / 2)
        ctx.arc(d, h - d, r, pi / 2, pi)
        ctx.arc(d, d, r, pi, pi * 3 / 2)
        ctx.close_path()

    w = pixbuf.get_width() + width * 2
    h = pixbuf.get_height() + width * 2
    r = min(radius, min(w, h) / 2)

    surface = cairo.ImageSurface(cairo.FORMAT_ARGB32, w, h)
    ctx = cairo.Context(surface)

    new_border_path(ctx, w, h, r + width)
    ctx.clip()
    ctx.set_source_rgba(color.red, color.green, color.blue, color.alpha)
    ctx.paint()

    new_border_path(ctx, w, h, r, width)
    Gdk.cairo_set_source_pixbuf(ctx, pixbuf, width, width)
    ctx.fill()

    return Gdk.pixbuf_get_from_surface(surface, 0, 0, w, h)


def add_border_widget(pixbuf, widget):
    """Like add_border() but uses the widget to get a border color and a
    border width.
    """

    context = widget.get_style_context()
    color = context.get_color(context.get_state())
    color.alpha *= 0.1
    scale_factor = widget.get_scale_factor()
    border_radius = get_border_radius() * scale_factor

    return add_border(pixbuf, color, width=scale_factor, radius=border_radius)


def scale(pixbuf, boundary, scale_up=True, force_copy=False):
    """Scale a pixbuf so it fits into the boundary.
    (preserves image aspect ratio)

    If `scale_up` is True, the resulting pixbuf can be larger than
    the original one.

    If `force_copy` is False the resulting pixbuf might be the passed one.

    Can not fail.
    """

    size = pixbuf.get_width(), pixbuf.get_height()

    scale_w, scale_h = calc_scale_size(boundary, size, scale_up)

    if (scale_w, scale_h) == size:
        if force_copy:
            return pixbuf.copy()
        return pixbuf

    return pixbuf.scale_simple(scale_w, scale_h, GdkPixbuf.InterpType.BILINEAR)


def calc_scale_size(boundary, size, scale_up=True):
    """Returns the biggest possible size to fit into the boundary,
    respecting the aspect ratio.

    If `scale_up` is True the result can be larger than size.

    All sizes have to be > 0.
    """

    bwidth, bheight = boundary
    iwidth, iheight = size

    if bwidth <= 0 or bheight <= 0 or iwidth <= 0 or iheight <= 0:
        raise ValueError

    scale_w, scale_h = iwidth, iheight

    if iwidth > bwidth or iheight > bheight or scale_up:
        bratio = float(bwidth) / bheight
        iratio = float(iwidth) / iheight

        if iratio > bratio:
            scale_w = bwidth
            scale_h = int(bwidth / iratio)
        else:
            scale_w = int(bheight * iratio)
            scale_h = bheight

    return scale_w, scale_h


def pixbuf_from_file(fileobj, boundary, scale_factor=1):
    """Returns a pixbuf with the maximum size defined by boundary.

    Can raise GLib.GError and return None
    """

    try:
        pixbuf = GdkPixbuf.Pixbuf.new_from_file(fileobj.name)
    except GLib.GError:
        try:
            loader = GdkPixbuf.PixbufLoader()
            loader.write(fileobj.read())
            loader.close()
            fileobj.seek(0, 0)
            pixbuf = loader.get_pixbuf()
        except OSError:
            return None

    w, h = boundary
    w *= scale_factor
    h *= scale_factor

    return scale(pixbuf, (w, h), scale_up=True)



# Copyright 2004-2006 Joe Wreschnig, Michael Urman, Niklas Janlert
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

from mutagen.trueaudio import TrueAudio

from ._id3 import ID3File


class TrueAudioFile(ID3File):
    format = "True Audio"
    mimes = ["audio/x-tta"]
    Kind = TrueAudio

    def _parse_info(self, info):
        self["~#length"] = info.length
        self["~#samplerate"] = info.sample_rate


loader = TrueAudioFile
types = [TrueAudioFile]
extensions = [".tta"]



# Copyright 2011, 2014 Christoph Reiter
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

from gi.repository import Gtk

from . import add_css


class MenuButton(Gtk.MenuButton):
    """TODO: remove. This used to be an implementation of Gtk.MenuButton
    when it wasn't available in gtk+
    """

    def __init__(self, widget=None, arrow=False, down=True):
        super().__init__()

        bbox = Gtk.HBox(spacing=3)
        if widget:
            bbox.pack_start(widget, True, True, 0)
        if arrow:
            arrow_type = Gtk.ArrowType.DOWN if down else Gtk.ArrowType.UP
            bbox.pack_start(Gtk.Arrow.new(arrow_type, Gtk.ShadowType.IN), True, True, 0)

        self.add(bbox)
        self.set_direction(Gtk.ArrowType.DOWN if down else Gtk.ArrowType.UP)

    def get_menu(self):
        return self.get_popup()

    def set_menu(self, menu):
        self.set_popup(menu)


class SmallMenuButton(MenuButton):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.set_size_request(26, 26)
        add_css(
            self,
            """
            * {
                padding: 0px 4px;
            }
        """,
        )



# Copyright 2011,2013 Christoph Reiter
#                2016 Nick Boultbee
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

from gi.repository import Gtk, Gst

if not Gst.ElementFactory.find("chromaprint"):
    from quodlibet import plugins

    raise plugins.MissingGstreamerElementPluginError("chromaprint")

from .submit import FingerprintDialog
from .util import get_api_key

from quodlibet import _
from quodlibet import config
from quodlibet import util
from quodlibet.qltk import Button, Frame, Icons
from quodlibet.qltk.entry import UndoEntry
from quodlibet.qltk.msg import ErrorMessage
from quodlibet.plugins.songsmenu import SongsMenuPlugin
from quodlibet.plugins.songshelpers import is_writable, is_finite, each_song


class AcoustidSearch(SongsMenuPlugin):
    PLUGIN_ID = "AcoustidSearch"
    PLUGIN_NAME = _("Acoustic Fingerprint Lookup")
    PLUGIN_DESC = _("Looks up song metadata through acoustic fingerprinting.")
    PLUGIN_ICON = Icons.NETWORK_WORKGROUP

    plugin_handles = each_song(is_finite, is_writable)

    def plugin_songs(self, songs):
        from .search import SearchWindow

        window = SearchWindow(songs, title=self.PLUGIN_NAME)
        window.show()

        # plugin_done checks for metadata changes and opens the write dialog
        window.connect("destroy", self.__plugin_done)

    def __plugin_done(self, *args):
        self.plugin_finish()


class AcoustidSubmit(SongsMenuPlugin):
    PLUGIN_ID = "AcoustidSubmit"
    PLUGIN_NAME = _("Submit Acoustic Fingerprints")
    PLUGIN_DESC = _(
        "Generates acoustic fingerprints using chromaprint "
        "and submits them to acoustid.org."
    )
    PLUGIN_ICON = Icons.NETWORK_WORKGROUP

    plugin_handles = each_song(is_finite, is_writable)

    def plugin_songs(self, songs):
        if not get_api_key():
            ErrorMessage(
                self,
                _("API Key Missing"),
                _(
                    "You have to specify an acoustid.org API key in the plugin "
                    "preferences before you can submit fingerprints."
                ),
            ).run()
        else:
            FingerprintDialog(songs)

    @classmethod
    def PluginPreferences(cls, win):
        box = Gtk.VBox(spacing=12)

        # api key section
        def key_changed(entry, *args):
            config.set("plugins", "fingerprint_acoustid_api_key", entry.get_text())

        button = Button(_("Request API key"), Icons.NETWORK_WORKGROUP)
        button.connect(
            "clicked", lambda s: util.website("https://acoustid.org/api-key")
        )
        key_box = Gtk.HBox(spacing=6)
        entry = UndoEntry()
        entry.set_text(get_api_key())
        entry.connect("changed", key_changed)
        label = Gtk.Label(label=_("API _key:"))
        label.set_use_underline(True)
        label.set_mnemonic_widget(entry)
        key_box.pack_start(label, False, True, 0)
        key_box.pack_start(entry, True, True, 0)
        key_box.pack_start(button, False, True, 0)

        box.pack_start(Frame(_("AcoustID Web Service"), child=key_box), True, True, 0)

        return box



# Copyright 2006 Joe Wreschnig
#        2016-23 Nick Boultbee
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

from typing import Any

from gi.repository import Gtk

from quodlibet import _, print_d


class Order:
    """Base class for all play orders

    In all methods:
        `playlist` is a GTK+ `ListStore` containing at least an `AudioFile`
                   as the first element in each row
                   (in the future there may be more elements per row).

        `iter`     is a `GtkTreeIter` for the song that just finished, if any.
                   If the song is not in the list, this iter will be `None`.
    """

    name: str | None = "unknown_order"
    """The name by which this order is known"""

    display_name = _("Unknown")
    """The (translated) display name"""

    accelerated_name = _("_Unknown")
    """The (translated) display name with (optional) accelerators"""

    replaygain_profiles = ["track"]
    """The ReplayGain mode(s) to use with this order.
    Shuffled ones typically prefer track modes"""

    priority = 100
    """The priority relative to other orders of its type.
    Larger numbers typically appear lower in lists."""

    def __init__(self):
        """Must have a zero-arg constructor"""

    def next(self, playlist, iter):
        """Not called directly, but the default implementation of
        `next_explicit` and `next_implicit` both just call this."""
        raise NotImplementedError

    def previous(self, playlist, iter):
        """Not called directly, but the default implementation of
        `previous_explicit` calls this.
        Right now there is no such thing as `previous_implicit`."""
        raise NotImplementedError

    def set(self, playlist, iter):
        """Not called directly, but the default implementations of
        `set_explicit` and `set_implicit` call this."""
        return iter

    def next_explicit(self, playlist, iter):
        """Not called directly, but the default implementations of
        `set_explicit` and `set_implicit` call this."""
        return self.next(playlist, iter)

    def next_implicit(self, playlist, iter):
        """Called when a song ends passively, e.g. it plays through."""
        return self.next(playlist, iter)

    def previous_explicit(self, playlist, iter):
        """Called when the user presses a "Previous" button."""
        return self.previous(playlist, iter)

    def set_explicit(self, playlist, iter):
        """Called when the user manually selects a song (at `iter`).
        If desired the play order can override that, or just
        log it and return the iter again.
        Note that playlist.current_iter is the current iter, if any.

        If the play order returns `None`,
        no action will be taken by the player.
        """
        return self.set(playlist, iter)

    def set_implicit(self, playlist, iter):
        """Called when the song is set by a means other than the UI."""
        return self.set(playlist, iter)

    def reset(self, playlist):
        """Called when there is no song ready to prepare for a new order.
        Implementations should reset the state of the current order,
        e.g. forgetting history / clearing pre-cached orders."""

    def __str__(self):
        """By default there is no interesting state"""
        return f"<{self.display_name}>"


class OrderRemembered(Order):
    """Shared class for all the shuffle modes that keep a memory
    of their previously played songs."""

    _played: list[Gtk.TreeIter]

    def __init__(self):
        super().__init__()
        self._played = []

    def next(self, playlist, iter):
        if iter is not None:
            self._played.append(iter)

    def previous(self, playlist, iter):
        if self._played:
            return self._played.pop()
        return None

    def set(self, playlist, iter):
        if iter is not None:
            self._played.append(iter)
        return iter

    def reset(self, playlist):
        del self._played[:]

    def remaining(self, playlist) -> dict[int, Any]:
        """Gets a map of all song indices to their song from the `playlist`
        that haven't yet been played"""

        def get_index(iter) -> int:
            path = playlist.get_path(iter)
            return path.get_indices()[0]

        played = set(map(get_index, self._played))
        print_d("Played %d of %d song(s)" % (len(self._played), len(playlist)))
        remaining = ((get_index(iter), value) for iter, value in playlist.iterrows())
        return {index: song for (index, song) in remaining if index not in played}


class OrderInOrder(Order):
    """Keep to the order of the supplied playlist"""

    name: str | None = "in_order"
    display_name = _("In Order")
    accelerated_name = _("_In Order")
    replaygain_profiles = ["album", "track"]
    priority = 0

    def next(self, playlist, iter):
        if iter is None:
            return playlist.get_iter_first()
        return playlist.iter_next(iter)

    def previous(self, playlist, iter):
        if len(playlist) == 0:
            return None
        if iter is None:
            return playlist[(len(playlist) - 1,)].iter
        path = max(1, playlist.get_path(iter).get_indices()[0])
        try:
            return playlist.get_iter((path - 1,))
        except ValueError:
            return None



# Copyright 2005 Joe Wreschnig
#           2012 Christoph Reiter
#      2011-2022 Nick Boultbee
#           2014 Jan Path
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

from gi.repository import Gtk, Pango, GLib, Gio
from senf import fsnative, fsn2text

from quodlibet.util.string.date import format_date
from quodlibet import _, print_d
from quodlibet import util
from quodlibet import config
from quodlibet import app
from quodlibet.pattern import Pattern
from quodlibet.qltk.views import TreeViewColumnButton
from quodlibet.qltk import add_css
from quodlibet.util.path import unexpand
from quodlibet.formats._audio import FILESYSTEM_TAGS
from quodlibet.qltk.x import CellRendererPixbuf


def create_songlist_column(model: Gtk.TreeModel, t):
    """Returns a SongListColumn instance for the given tag"""

    if t in ["~#added", "~#mtime", "~#lastplayed", "~#laststarted"]:
        return DateColumn(t)
    if t in ["~length", "~#length"]:
        return LengthColumn()
    if t == "~#filesize":
        return FilesizeColumn()
    if t in ["~rating"]:
        return RatingColumn()
    if t.startswith("~#"):
        return NumericColumn(t)
    if t in FILESYSTEM_TAGS:
        return FSColumn(t)
    if "<" in t:
        return PatternColumn(t)
    if "~" not in t and t != "title":
        return NonSynthTextColumn(model, t)
    return WideTextColumn(t)


def _highlight_current_cell(cr, background_area, cell_area, flags):
    """Draws a 'highlighting' background for the cell. Look depends on
    the active theme.
    """

    # Use drawing code/CSS for Entry (reason being that it looks best here)
    dummy_widget = Gtk.Entry()
    style_context = dummy_widget.get_style_context()
    style_context.save()
    # Make it less prominent
    state = Gtk.StateFlags.INSENSITIVE | Gtk.StateFlags.BACKDROP
    style_context.set_state(state)
    color = style_context.get_border_color(state)
    add_css(
        dummy_widget,
        "* { border-color: rgba(%d, %d, %d, 0.3); }"
        % (color.red * 255, color.green * 255, color.blue * 255),
    )
    ba = background_area
    ca = cell_area
    # Draw over the left and right border so we don't see the rounded corners
    # and borders. Use height for the overshoot as rounded corners + border
    # should never be larger than the height..
    # Ideally we would draw over the whole background but the cell area only
    # redraws the cell_area so we get leftover artifacts if we draw
    # above/below.
    draw_area = (ba.x - ca.height, ca.y, ba.width + ca.height * 2, ca.height)
    cr.save()
    cr.new_path()
    cr.rectangle(ba.x, ca.y, ba.width, ca.height)
    cr.clip()
    Gtk.render_background(style_context, cr, *draw_area)
    Gtk.render_frame(style_context, cr, *draw_area)
    cr.restore()
    style_context.restore()


class SongListCellAreaBox(Gtk.CellAreaBox):
    highlight = False

    def do_render(
        self, context, widget, cr, background_area, cell_area, flags, paint_focus
    ):
        if self.highlight and not flags & Gtk.CellRendererState.SELECTED:
            _highlight_current_cell(cr, background_area, cell_area, flags)
        return Gtk.CellAreaBox.do_render(
            self, context, widget, cr, background_area, cell_area, flags, paint_focus
        )

    def do_apply_attributes(self, tree_model, iter_, is_expander, is_expanded):
        self.highlight = tree_model.get_path(iter_) == tree_model.current_path
        return Gtk.CellAreaBox.do_apply_attributes(
            self, tree_model, iter_, is_expander, is_expanded
        )


class SongListColumn(TreeViewColumnButton):
    can_edit = False
    """Whether this column can support editing"""

    def __init__(self, tag):
        """tag e.g. 'artist'"""

        title = self._format_title(tag)
        super().__init__(title=title, cell_area=SongListCellAreaBox())
        self.set_tooltip_text(title)
        self.header_name = tag

        self.set_sizing(Gtk.TreeViewColumnSizing.FIXED)
        self.set_visible(True)
        self.set_sort_indicator(False)

        self._last_rendered = None

    def _format_title(self, tag):
        """Format the column title based on the tag"""

        return util.tag(tag)

    def _needs_update(self, value):
        """Call to check if the last passed value was the same.

        This is used to reduce formatting if the input is the same
        either because of redraws or all columns have the same value
        """

        if self._last_rendered == value:
            return False
        self._last_rendered = value
        return True


class TextColumn(SongListColumn):
    """Base text column"""

    def __init__(self, tag):
        super().__init__(tag)

        self._render = Gtk.CellRendererText()
        self.pack_start(self._render, True)
        self.set_cell_data_func(self._render, self._cdf)
        self.set_clickable(True)

        # We check once in a while if the font size has changed. If it has
        # we reset the min/fixed width and force at least one cell to update
        # (which might trigger other column size changes..)
        self._last_width = None
        self._force_update = False
        self._deferred_width_check = util.DeferredSignal(
            self._check_width_update, timeout=500
        )

        def on_tv_changed(column, old, new):
            if new is None:
                self._deferred_width_check.abort()
            else:
                self._deferred_width_check.call()

        self.connect("tree-view-changed", on_tv_changed)

    def _get_min_width(self):
        return -1

    def _cell_width(self, text):
        """Returns the column width needed for the passed text"""

        widget = self.get_tree_view()
        assert widget is not None
        layout = widget.create_pango_layout(text)
        text_width = layout.get_pixel_size()[0]
        cell_pad = self._render.get_property("xpad")

        return text_width + 8 + cell_pad

    def _check_width_update(self):
        width = self._cell_width("abc 123")
        if self._last_width == width:
            self._force_update = False
            return
        self._last_width = width
        self._force_update = True
        self.queue_resize()

    def _needs_update(self, value):
        return self._force_update or super()._needs_update(value)

    def _cdf(self, column, cell, model, iter_, user_data):
        self._deferred_width_check()
        if self._force_update:
            min_width = self._get_min_width()
            self.set_min_width(min_width)
            if not self.get_resizable():
                self.set_fixed_width(min_width)
            # calling it in the cell_data_func leads to broken drawing..
            GLib.idle_add(self.queue_resize)

        value = self._fetch_value(model, iter_)
        if not self._needs_update(value):
            return
        self._apply_value(model, iter_, cell, value)

    def _fetch_value(self, model, iter_):
        """Should return everything needed for formatting the final value"""

        raise NotImplementedError

    def _apply_value(self, model, iter_, cell, value):
        """Should format the value and set it on the cell renderer"""

        raise NotImplementedError


class RatingColumn(TextColumn):
    """Render ~rating directly

    (simplifies filtering, saves a function call).
    """

    def __init__(self, *args, **kwargs):
        super().__init__("~rating", *args, **kwargs)
        self.set_expand(False)
        self.set_resizable(False)

    def _get_min_width(self):
        return self._cell_width(util.format_rating(1.0))

    def _fetch_value(self, model, iter_):
        song = model.get_value(iter_)
        rating = song.get("~#rating")
        default = config.RATINGS.default
        return (rating, default)

    def _apply_value(self, model, iter_, cell, value):
        rating, default = value
        cell.set_sensitive(rating is not None)
        value = rating if rating is not None else default
        cell.set_property("text", util.format_rating(value))


class WideTextColumn(TextColumn):
    """Resizable and ellipsized at the end. Used for any key with
    a '~' in it, and 'title'.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._render.set_property("ellipsize", Pango.EllipsizeMode.END)
        self.set_resizable(True)

    def _get_min_width(self):
        return self._cell_width("000")

    def _fetch_value(self, model, iter_):
        return model.get_value(iter_).comma(self.header_name)

    def _apply_value(self, model, iter_, cell, value):
        cell.set_property("text", value)


class DateColumn(WideTextColumn):
    """The '~#' keys that are dates."""

    def _fetch_value(self, model, iter_):
        return model.get_value(iter_)(self.header_name)

    def _apply_value(self, model, iter_, cell, stamp):
        if not stamp:
            cell.set_property("text", _("Never"))
        else:
            fmt = config.gettext("settings", "datecolumn_timestamp_format")
            text = format_date(stamp, fmt)
            cell.set_property("text", text)


class NonSynthTextColumn(WideTextColumn):
    """Optimize for non-synthesized keys by grabbing them directly.
    Used for any tag without a '~' except 'title'.
    """

    can_edit = True

    def __row_edited(self, render, path, new: str, model: Gtk.TreeModel) -> None:
        print_d(f"Trying to edit {self.header_name} to {new!r}")
        model[path][0][self.header_name] = new
        model.path_changed(path)

    def __init__(self, model, tag):
        super().__init__(tag)
        self._render.connect("edited", self.__row_edited, model)

    def _fetch_value(self, model, iter_):
        return model.get_value(iter_).get(self.header_name, "")

    def _apply_value(self, model, iter_, cell, value):
        cell.set_property("text", value.replace("\n", ", "))


class FSColumn(WideTextColumn):
    """Contains text in the filesystem encoding, so needs to be
    decoded safely (and also more slowly).
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._render.set_property("ellipsize", Pango.EllipsizeMode.MIDDLE)

    def _fetch_value(self, model, iter_):
        values = model.get_value(iter_).list(self.header_name)
        return values[0] if values else fsnative("")

    def _apply_value(self, model, iter_, cell, value):
        cell.set_property("text", fsn2text(unexpand(value)))


class PatternColumn(WideTextColumn):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        try:
            self._pattern = Pattern(self.header_name)
        except ValueError:
            self._pattern = None

    def _format_title(self, tag):
        return util.pattern(tag)

    def _fetch_value(self, model, iter_):
        song = model.get_value(iter_)
        if self._pattern is not None:
            return self._pattern % song
        return ""

    def _apply_value(self, model, iter_, cell, value):
        cell.set_property("text", value)


class NumericColumn(TextColumn):
    """Any '~#' keys except dates."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._render.set_property("xalign", 1.0)
        self.set_alignment(1.0)
        self.set_expand(False)
        self.set_resizable(False)

        self._texts = {}
        self._timeout = None

        def on_tv_changed(column, old, new):
            if new is None and self._timeout is not None:
                GLib.source_remove(self._timeout)
                self._timeout = None

        self.connect("tree-view-changed", on_tv_changed)

    def _get_min_width(self):
        """Give the initial and minimum width. override if needed"""

        # Best efforts for the general minimum width case
        # Allows well for >=1000 Kbps, -12.34 dB RG values, "Length" etc
        return self._cell_width("-22.22")

    def _fetch_value(self, model, iter_):
        return model.get_value(iter_).comma(self.header_name)

    def _apply_value(self, model, iter_, cell, value):
        if isinstance(value, float):
            text = f"{round(value, 2):.2f}"
        else:
            text = str(value)

        cell.set_property("text", text)
        self._recalc_width(model.get_path(iter_), text)

    def _delayed_recalc(self):
        self._timeout = None

        tv = self.get_tree_view()
        assert tv is not None
        range_ = tv.get_visible_range()
        if not range_:
            return

        start, end = range_
        start = start[0]
        end = end[0]

        # compute the cell width for all drawn cells in range +/- 3
        for key, value in list(self._texts.items()):
            if not (start - 3) <= key <= (end + 3):
                del self._texts[key]
            elif isinstance(value, str):
                self._texts[key] = self._cell_width(value)

        # resize if too small or way too big and above the minimum
        width = self.get_width()
        needed_width = max([self._get_min_width()] + list(self._texts.values()))
        if width < needed_width:
            self._resize(needed_width)
        elif width - needed_width >= self._cell_width("0"):
            self._resize(needed_width)

    def _resize(self, width):
        # In case the treeview has no other expanding columns, setting the
        # width will have no effect on the actual width. Calling queue_resize()
        # in that case would result in an endless recalc loop. So stop here.
        if width == self.get_fixed_width() and width == self.get_max_width():
            return

        self.set_fixed_width(width)
        self.set_max_width(width)
        self.queue_resize()

    def _recalc_width(self, path, text):
        self._texts[path[0]] = text
        if self._timeout is not None:
            GLib.source_remove(self._timeout)
            self._timeout = None
        self._timeout = GLib.idle_add(self._delayed_recalc, priority=GLib.PRIORITY_LOW)


class LengthColumn(NumericColumn):
    def __init__(self):
        super().__init__("~#length")

    def _get_min_width(self):
        # 1:22:22, allows entire albums as files (< 75mins)
        return self._cell_width(util.format_time_display(60 * 82 + 22))

    def _fetch_value(self, model, iter_):
        return model.get_value(iter_).get("~#length", 0)

    def _apply_value(self, model, iter_, cell, value):
        text = util.format_time_display(value)
        cell.set_property("text", text)
        self._recalc_width(model.get_path(iter_), text)


class FilesizeColumn(NumericColumn):
    def __init__(self):
        super().__init__("~#filesize")

    def _get_min_width(self):
        # e.g "2.22 MB"
        return self._cell_width(util.format_size(2.22 * (1024**2)))

    def _fetch_value(self, model, iter_):
        return model.get_value(iter_).get("~#filesize", 0)

    def _apply_value(self, model, iter_, cell, value):
        text = util.format_size(value)
        cell.set_property("text", text)
        self._recalc_width(model.get_path(iter_), text)


class CurrentColumn(SongListColumn):
    """Displays the current song indicator, either a play or pause icon."""

    def __init__(self):
        super().__init__("~current")
        self._render = CellRendererPixbuf()
        self.pack_start(self._render, True)
        self._render.set_property("xalign", 0.5)

        self.set_fixed_width(24)
        self.set_expand(False)
        self.set_cell_data_func(self._render, self._cdf)

    def _format_title(self, tag):
        return ""

    def _cdf(self, column, cell, model, iter_, user_data):
        PLAY = "media-playback-start"
        PAUSE = "media-playback-pause"
        STOP = "media-playback-stop"
        ERROR = "dialog-error"

        row = model[iter_]

        if row.path == model.current_path:
            player = app.player
            if player.error:
                name = ERROR
            elif model.sourced:
                name = [PLAY, PAUSE][player.paused]
            else:
                name = STOP
        else:
            name = None

        if not self._needs_update(name):
            return

        if name is not None:
            gicon = Gio.ThemedIcon.new_from_names([name + "-symbolic", name])
        else:
            gicon = None

        cell.set_property("gicon", gicon)



# Copyright 2008 Joe Wreschnig
#      2016-2025 Nick Boultbee
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
import functools

from quodlibet import _
from quodlibet.plugins.songshelpers import any_song, is_finite, is_writable
from quodlibet.plugins.songsmenu import SongsMenuPlugin
from quodlibet.qltk import Icons
from quodlibet.util.collection import PEOPLE
from quodlibet.util.string.filter import remove_diacritics

SUPPORTED_TAGS = ["album", "artist", "albumartist", "composer", "performer"]

# Pretty arbitrary, but helps a lot IMO
FAMOUS_COMPOSERS = {
    "bach",
    "mozart",
    "beethoven",
    "chopin",
    "brahms",
    "faure",
    "rachmaninov",
    "rachmaninoff",
    "puccini",
    "mendelssohn",
    "liszt",
    "elgar",
    "mahler",
    "debussy",
    "bartok",
    "schumann",
    "dvorak",
    "holst",
    "britten",
    "straus",
    "williams",
    "zimmer",
}


def artist_to_sort(tag: str, artist: str) -> str | None:
    try:
        first, rest = artist.split(" ", 1)
    except ValueError:
        pass
    else:
        # "The Beach Boys" -> "Beach Boys, The" etc
        if first.lower() in ["a", "the"]:
            return f"{rest}, {first}"
        # We really want "Amadeus Mozart" -> "Mozart, Amadeus"
        # But not really "Rage Against The Machine" -> "Machine, Rage Against The"
        # So let's restrict to composer then.
        try:
            rest, last = artist.rsplit(" ", 1)
            last_normalised = remove_diacritics(last).lower()
            if tag in {"composer"} or last_normalised in FAMOUS_COMPOSERS:
                return f"{last}, {rest}"
        except ValueError:
            pass
    return None


def album_to_sort(album: str) -> str | None:
    try:
        first, rest = album.split(" ", 1)
    except ValueError:
        pass
    else:
        if first.lower() in ["a", "the"]:
            return f"{rest}, {first}"
    return None


class MakeSortTags(SongsMenuPlugin):
    PLUGIN_ID = "SortTags"
    PLUGIN_NAME = _("Create Sort Tags")
    PLUGIN_DESC_MARKUP = (
        _("Guesses sort tags for albums and people.") + "\n\n" + _("e.g.") + "\n\n "
        "<tt>album</tt>: <i>The Greatest Hits</i> → <i>Greatest Hits, The</i>"
        "<tt>composer</tt>: <i>Irving Berlin</i> → <i>Berlin, Irving</i>\n"
        "<tt>artist</tt>: <i>Franz Liszt</i> → <i>Liszt, Franz</i>\n"
        "<tt>artist</tt>: <i>The Beach Boys</i> → <i>Beach Boys, The</i>\n\n"
        + _("Applies to tags: <tt>%s</tt>.")
        % ", ".join(SUPPORTED_TAGS)
    )
    PLUGIN_ICON = Icons.EDIT

    plugin_handles = any_song(is_writable, is_finite)

    def plugin_song(self, song):
        for tag in SUPPORTED_TAGS:
            func = (
                functools.partial(artist_to_sort, tag)
                if tag in PEOPLE
                else album_to_sort
            )
            values = [v for tag in song.list(tag) if (v := func(tag))]
            if values and (tag + "sort") not in song:
                song[tag + "sort"] = "\n".join(values)



# Copyright 2006 Joe Wreschnig
#      2012-2022 Nick Boultbee
#           2014 Christoph Reiter
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

"""
Librarians for libraries.
"""

import itertools
from collections.abc import Iterable, Iterator, Generator

from gi.repository import GObject

from quodlibet.library.base import Library
from quodlibet.library.playlist import PlaylistLibrary
from quodlibet.util.dprint import print_d, print_w
from senf import fsnative


class Librarian(GObject.GObject):
    """The librarian is a nice interface to all active libraries.

    Librarians are a kind of meta-library. When any of their
    registered libraries fire a signal, they fire the same
    signal. Likewise, they provide various methods equivalent to the
    ones found in libraries that group the results of the real
    libraries.

    Attributes:
    libraries -- a dict mapping library names to libraries
    """

    __gsignals__ = {
        "changed": (GObject.SignalFlags.RUN_LAST, None, (object,)),
        "removed": (GObject.SignalFlags.RUN_LAST, None, (object,)),
        "added": (GObject.SignalFlags.RUN_LAST, None, (object,)),
    }

    def __init__(self):
        super().__init__()
        self.libraries: dict[str, Library] = {}
        self.__signals = {}

    def destroy(self) -> None:
        pass

    def register(self, library: Library, name: str) -> None:
        """Register a library with this librarian."""
        if name in self.libraries or name in self.__signals:
            raise ValueError(f"library {name!r} is already active")

        added_sig = library.connect("added", self.__added)
        removed_sig = library.connect("removed", self.__removed)
        changed_sig = library.connect("changed", self.__changed)
        self.libraries[name] = library
        self.__signals[library] = [added_sig, removed_sig, changed_sig]

    def _unregister(self, library: Library, name: str) -> None:
        # This function, unlike register, should be private.
        # Libraries get unregistered at the discretion of the
        # librarian, not the libraries.
        del self.libraries[name]
        for signal_id in self.__signals[library]:
            library.disconnect(signal_id)
        del self.__signals[library]

    # FIXME: We can be smarter about this -- queue a list of items
    # and fire the signal after a short wait, to take advantage of
    # a case where many libraries fire a signal at the same time (or
    # one fires a signal often).

    def __changed(self, _library: Library, items: Iterable) -> None:
        self.emit("changed", items)

    def __added(self, _library: Library, items: Iterable) -> None:
        self.emit("added", items)

    def __removed(self, _library: Library, items: Iterable) -> None:
        self.emit("removed", items)

    def changed(self, items: Iterable) -> None:
        """Triage the items and inform their real libraries."""

        for library in self.libraries.values():
            in_library = {item for item in items if item in library}
            if in_library:
                library._changed(in_library)

    def __getitem__(self, key):
        """Find a item given its key."""
        for library in self.libraries.values():
            try:
                return library[key]
            except KeyError:
                pass
        else:
            raise KeyError(key)

    def get(self, key, default=None):
        try:
            return self[key]
        except KeyError:
            return default

    def remove(self, items: Iterable):
        """Remove items from all libraries."""
        for library in self.libraries.values():
            library.remove(items)

    def __contains__(self, item) -> bool:
        """Check if a key or item is in the library."""
        for library in self.libraries.values():
            if item in library:
                return True
        else:
            return False

    def __iter__(self) -> Iterator:
        """Iterate over all items in all libraries."""
        return itertools.chain(*self.libraries.values())

    def move(self, items: Iterable, from_: Library, to: Library) -> None:
        """Move items from one library to another.

        This causes 'removed' signals on the from library, and 'added'
        signals on the 'to' library, but will not cause any signals
        to be emitted via this librarian.
        """
        try:
            from_.handler_block(self.__signals[from_][1])
            to.handler_block(self.__signals[to][0])
            from_.remove(items)
            to.add(items)
        finally:
            from_.handler_unblock(self.__signals[from_][1])
            to.handler_unblock(self.__signals[to][0])

    def move_root(self, old_root: fsnative, new_root: fsnative, **kwargs) -> Generator:
        if old_root == new_root:
            print_d("Not moving to same root")
            return
        for library in self.libraries.values():
            if hasattr(library, "move_root"):
                yield from library.move_root(old_root, new_root, **kwargs)

    def remove_roots(self, old_roots: Iterable[fsnative]) -> Generator:
        if not old_roots:
            return
        for library in self.libraries.values():
            if hasattr(library, "remove_roots"):
                yield from library.remove_roots(old_roots)


class SongLibrarian(Librarian):
    """A librarian for SongLibraries."""

    def tag_values(self, tag):
        """Return a set of all values for the given tag."""
        return {
            value for lib in self.libraries.values() for value in lib.tag_values(tag)
        }

    def rename(self, song, newname, changed=None):
        """Rename the song in all libraries it belongs to.

        The 'changed' signal will fire for any library the song is in
        except if a set() is passed as changed.
        """
        # This needs to poke around inside the library directly.  If
        # it uses add/remove to handle the songs it fires incorrect
        # signals. If it uses the library's rename method, it breaks
        # the call for future libraries because the item's key has
        # changed. So, it needs to reimplement the method.
        re_add = []
        print_d(f"Renaming {song.key!r} to {newname!r}")
        for library in self.libraries.values():
            try:
                del library._contents[song.key]
            except KeyError:
                pass
            else:
                re_add.append(library)
        song.rename(newname)
        for library in re_add:
            library._contents[song.key] = song
            if changed is None:
                library._changed({song})
            else:
                changed.add(song)

    def reload(self, item, changed=None, removed=None):
        """Reload a song (for all libraries), possibly noting its status.

        If sets are given, it assumes the caller will handle signals,
        and only updates the sets. Otherwise, it handles signals
        itself. It *always* handles library contents, so do not
        try to remove (again) a song that appears in the removed set.
        """

        had_item = []
        print_d(f"Reloading {item.key!r}")
        for library in self.libraries.values():
            try:
                del library._contents[item.key]
            except KeyError:
                pass
            else:
                had_item.append(library)
        try:
            library = had_item[0]
        except IndexError:
            return

        # Rely on the first library in the list to do the actual
        # load, then just inform the other libraries what happened.
        was_changed, was_removed = library._load_item(item, force=True)
        assert not (was_changed and was_removed)

        if was_removed:
            if removed is None:
                for library in had_item:
                    library.emit("removed", {item})
            else:
                removed.add(item)
        elif was_changed:
            for library in had_item:
                library._contents[item.key] = item

            if changed is None:
                for library in had_item:
                    library.emit("changed", {item})
            else:
                changed.add(item)

    @property
    def playlists(self):
        for lib in self.libraries.values():
            if isinstance(lib, PlaylistLibrary):
                return lib
            try:
                return lib.playlists
            except AttributeError:
                pass
        print_w(f"No playlist library found: {self.libraries}")
        raise ValueError("No playlists library found")



# Copyright 2012,2014 Christoph Reiter
#           2014,2017 Nick Boultbee
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

import sys
import os

from gi.repository import Gtk, Gdk

from quodlibet import config
from quodlibet.qltk import get_top_parent, is_wayland, gtk_version, is_accel
from quodlibet.qltk.x import Button
from quodlibet.util import DeferredSignal, print_d, print_w
from quodlibet.util import connect_obj, connect_destroy


def on_first_map(window, callback, *args, **kwargs):
    """Calls callback when the passed Gtk.Window is first visible
    on screen or it already is.
    """

    assert isinstance(window, Gtk.Window)

    if window.get_mapped():
        callback(*args, **kwargs)
        return False

    id_ = [0]

    def on_map(*otherargs):
        window.disconnect(id_[0])
        callback(*args, **kwargs)

    id_[0] = window.connect("map", on_map)

    return False


def should_use_header_bar():
    settings = Gtk.Settings.get_default()
    if not settings:
        return False
    if not hasattr(settings.props, "gtk_dialogs_use_header"):
        return False
    return settings.get_property("gtk-dialogs-use-header")


def fix_default_size(width, height):
    # https://bugzilla.gnome.org/show_bug.cgi?id=740922
    if gtk_version < (3, 19):
        # fixed with 3.20:
        #   https://bugzilla.gnome.org/show_bug.cgi?id=756618
        if width != -1:
            width += min((width - 174), 56)
        if height != -1:
            height += 84
    return (width, height)


class Dialog(Gtk.Dialog):
    """A Gtk.Dialog subclass which supports the use_header_bar property
    for all Gtk versions and will ignore it if header bars shouldn't be
    used according to GtkSettings.
    """

    def __init__(self, *args, **kwargs):
        if not should_use_header_bar():
            kwargs.pop("use_header_bar", None)
        super().__init__(*args, **kwargs)

    def get_titlebar(self):
        try:
            # gtk+ >=3.16
            return super().get_titlebar()
        except AttributeError:
            return None

    def set_default_size(self, width, height):
        if self.get_titlebar():
            width, height = fix_default_size(width, height)
        else:
            # In case we don't use a headerbar we have to add an additional
            # row of buttons in the content box. To get roughly the same
            # content height make the window a bit taller.
            if height != -1:
                height += 20
        super().set_default_size(width, height)

    def add_icon_button(self, label, icon_name, response_id):
        """Like add_button() but allows to pass an icon name"""

        button = Button(label, icon_name)
        # file chooser uses grab_default() on this
        button.set_can_default(True)
        button.show()
        self.add_action_widget(button, response_id)
        return button


class Window(Gtk.Window):
    """Base window class the keeps track of all window instances.

    All active instances can be accessed through Window.windows.
    By defining dialog=True as a kwarg binds Escape to close, otherwise
    ^W will close the window.
    """

    windows: list[Gtk.Window] = []
    _preven_inital_show = False

    def __init__(self, *args, **kwargs):
        self._header_bar = None
        dialog = kwargs.pop("dialog", True)
        super().__init__(*args, **kwargs)
        type(self).windows.append(self)
        if dialog:
            # Modal is the only way to center the window on the parent
            # with wayland atm
            if is_wayland():
                self.set_modal(True)
            self.set_type_hint(Gdk.WindowTypeHint.DIALOG)
        self.set_destroy_with_parent(True)
        self.set_position(Gtk.WindowPosition.CENTER_ON_PARENT)
        connect_obj(self, "destroy", type(self).windows.remove, self)
        self.connect("key-press-event", self._on_key_press)

    def _on_key_press(self, widget, event):
        is_dialog = self.get_type_hint() == Gdk.WindowTypeHint.DIALOG

        if (is_dialog and is_accel(event, "Escape")) or (
            not is_dialog and is_accel(event, "<Primary>w")
        ):
            # Do not close the window if we edit a Gtk.CellRendererText.
            # Focus the treeview instead.
            if isinstance(self.get_focus(), Gtk.Entry) and isinstance(
                self.get_focus().get_parent(), Gtk.TreeView
            ):
                self.get_focus().get_parent().grab_focus()
                return Gdk.EVENT_PROPAGATE
            self.close()
            return Gdk.EVENT_STOP

        if not is_dialog and is_accel(event, "F11"):
            self.toggle_fullscreen()
            return Gdk.EVENT_STOP

        return Gdk.EVENT_PROPAGATE

    def toggle_fullscreen(self):
        """Toggle the fullscreen mode of the window depending on its current
        state. If the windows isn't realized it will switch to fullscreen
        when it does.
        """

        window = self.get_window()
        if not window:
            is_fullscreen = False
        else:
            is_fullscreen = window.get_state() & Gdk.WindowState.FULLSCREEN

        if is_fullscreen:
            self.unfullscreen()
        else:
            self.fullscreen()

    def set_default_size(self, width, height):
        if self._header_bar:
            width, height = fix_default_size(width, height)
        super().set_default_size(width, height)

    def use_header_bar(self):
        """Try to use a headerbar, returns the widget or None in case
        GTK+ is too old or headerbars are disabled (under xfce for example)
        """

        assert not self._header_bar

        if not should_use_header_bar():
            return False

        header_bar = Gtk.HeaderBar()
        header_bar.set_show_close_button(True)
        header_bar.show()
        old_title = self.get_title()
        self.set_titlebar(header_bar)
        if old_title is not None:
            self.set_title(old_title)
        self._header_bar = header_bar
        self.set_default_size(*self.get_default_size())
        return header_bar

    def has_close_button(self):
        """Returns True in case we are sure that the window decorations include
        a close button.
        """

        if self.get_type_hint() == Gdk.WindowTypeHint.NORMAL:
            return True

        if os.name == "nt":
            return True

        if sys.platform == "darwin":
            return True

        if self._header_bar is not None:
            return self._header_bar.get_show_close_button()

        return True

    def present(self):
        """A version of present that also works if not called from an event
        handler (there is no active input event).
        See https://bugzilla.gnome.org/show_bug.cgi?id=688830
        """

        try:
            from gi.repository import GdkX11
        except ImportError:
            super().present()
        else:
            window = self.get_window()
            if window and isinstance(window, GdkX11.X11Window):
                timestamp = GdkX11.x11_get_server_time(window)
                self.present_with_time(timestamp)
            else:
                super().present()

    def set_transient_for(self, parent):
        """Set a parent for the window.

        In case parent=None, fall back to the main window.

        """

        is_toplevel = parent and parent.props.type == Gtk.WindowType.TOPLEVEL

        if parent is None or not is_toplevel:
            if parent:
                print_w(f"Not a toplevel window set for: {self!r}")
            from quodlibet import app

            parent = app.window
        super().set_transient_for(parent)

    @classmethod
    def prevent_inital_show(cls, value):
        cls._preven_inital_show = bool(value)

    def show_maybe(self):
        """Show the window, except if prevent_inital_show() was called and
        this is the first time.

        Returns whether the window was shown.
        """

        if not self._preven_inital_show:
            self.show()
        return not self._preven_inital_show


class PersistentWindowMixin:
    """A mixin for saving/restoring window size/position/maximized state"""

    def enable_window_tracking(self, config_prefix, size_suffix=""):
        """Enable tracking/saving of changes and restore size/pos/maximized.

        Make sure to call set_transient_for() before since position is
        restored relative to the parent in this case.

        config_prefix -- prefix for the config key
                         (prefix_size, prefix_position, prefix_maximized)
        size_suffix -- optional suffix for saving the size. For cases where the
                       window has multiple states with different content sizes.
                       (example: edit tags with one song or multiple)

        """

        self.__state = 0
        self.__name = config_prefix
        self.__size_suffix = size_suffix
        self.__save_size_pos_deferred = DeferredSignal(
            self.__do_save_size_pos, timeout=50, owner=self
        )
        self.connect("configure-event", self.__configure_event)
        self.connect("window-state-event", self.__window_state_changed)
        self.connect("notify::visible", self.__visible_changed)
        parent = self.get_transient_for()
        if parent:
            connect_destroy(parent, "configure-event", self.__parent_configure_event)
        self.__restore_window_state()

    def __visible_changed(self, *args):
        if not self.get_visible():
            # https://bugzilla.gnome.org/show_bug.cgi?id=731287
            # if we restore after hide, mutter will remember for the next show
            # hurray!
            self.__restore_window_state()

    def __restore_window_state(self):
        if not is_wayland():
            self.__restore_state()
            self.__restore_position()
        self.__restore_size()

    def __conf(self, name):
        if name == "size":
            name += "_" + self.__size_suffix
        return f"{self.__name}_{name}"

    def __restore_state(self):
        print_d("Restore state")
        if config.getint("memory", self.__conf("maximized"), 0):
            self.maximize()
        else:
            self.unmaximize()

    def __restore_position(self):
        print_d("Restore position")
        pos = config.get("memory", self.__conf("position"), "")
        if not pos:
            return

        try:
            x, y = map(int, pos.split())
        except ValueError:
            return

        parent = self.get_transient_for()
        if parent:
            px, py = parent.get_position()
            x += px
            y += py

        self.move(x, y)

    def __restore_size(self):
        print_d("Restore size")
        value = config.get("memory", self.__conf("size"), "")
        if not value:
            return

        try:
            x, y = map(int, value.split())
        except ValueError:
            return

        screen = self.get_screen()
        x = min(x, screen.get_width())
        y = min(y, screen.get_height())
        if x >= 1 and y >= 1:
            self.resize(x, y)

    def __parent_configure_event(self, window, event):
        # since our position is relative to the parent if we have one,
        # we also need to save our position if the parent position changes
        self.__do_save_pos()
        return False

    def __configure_event(self, window, event):
        # xfwm4 resized the window before it maximizes it, which leads
        # to QL remembering the wrong size. Work around that by waiting
        # until configure-event settles down, at which point the maximized
        # state should be set

        self.__save_size_pos_deferred()
        return False

    def _should_ignore_state(self):
        if self.__state & Gdk.WindowState.MAXIMIZED:
            return True
        if self.__state & Gdk.WindowState.FULLSCREEN:
            return True
        if not self.get_visible():
            return True
        return False

    def __do_save_size_pos(self):
        if self._should_ignore_state():
            return

        width, height = self.get_size()
        value = "%d %d" % (width, height)
        config.set("memory", self.__conf("size"), value)

        self.__do_save_pos()

    def __do_save_pos(self):
        if self._should_ignore_state():
            return

        x, y = self.get_position()
        parent = self.get_transient_for()
        if parent:
            px, py = parent.get_position()
            x -= px
            y -= py

        pos_value = f"{x} {y}"
        config.set("memory", self.__conf("position"), pos_value)

    def __window_state_changed(self, window, event):
        self.__state = event.new_window_state
        if self.__state & Gdk.WindowState.WITHDRAWN:
            return
        maximized = int(self.__state & Gdk.WindowState.MAXIMIZED)
        config.set("memory", self.__conf("maximized"), maximized)


class _Unique:
    """A mixin for window-like classes to ensure one instance per class."""

    __window = None

    def __new__(cls, *args, **kwargs):
        window = cls.__window
        if window is None:
            return super().__new__(cls, *args, **kwargs)
        # Look for widgets in the args, if there is one and it has
        # a new top level window, re-parent and reposition the window.
        widgets = [w for w in args if isinstance(w, Gtk.Widget)]
        if widgets:
            parent = window.get_transient_for()
            new_parent = get_top_parent(widgets[0])
            if parent and new_parent and parent is not new_parent:
                window.set_transient_for(new_parent)
                window.hide()
                window.show()
        window.present()
        return window

    @classmethod
    def is_not_unique(cls):
        """Returns True if a window instance already exists."""
        return bool(cls.__window)

    def __init__(self, *args, **kwargs):
        if type(self).__window:
            return
        type(self).__window = self
        super().__init__(*args, **kwargs)
        connect_obj(self, "destroy", self.__destroy, self)

    def __destroy(self, *args):
        type(self).__window = None


class UniqueWindow(_Unique, Window):
    pass



# Copyright 2004-2005 Joe Wreschnig, Michael Urman
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

import mutagen.apev2

from quodlibet.util.path import get_temp_cover_file

from ._audio import AudioFile
from ._image import APICType, EmbeddedImage
from ._misc import AudioFileError, translate_errors


def get_cover_type(key, value):
    """Returns an APICType or None if the tag isn't an image"""

    if value.kind != mutagen.apev2.BINARY:
        return None

    type_map = {
        "cover art (front)": APICType.COVER_FRONT,
        "cover art (back)": APICType.COVER_BACK,
    }

    return type_map.get(key.lower())


def parse_cover(key, value):
    """Returns a EmbeddedImage or None"""

    # http://www.hydrogenaud.io/forums/index.php?showtopic=40603

    cover_type = get_cover_type(key, value)
    if cover_type is None:
        return None

    parts = value.value.split(b"\x00", 1)
    if len(parts) != 2:
        return None

    f = get_temp_cover_file(parts[-1])
    if not f:
        return None

    return EmbeddedImage(f, "image/", type_=cover_type)


def write_cover(image):
    """Takes EmbeddedImage and returns a (key, value) tuple

    Raises:
        AudioFileError
    """

    if image.type == APICType.COVER_FRONT:
        key = "Cover Art (Front)"
    else:
        key = "Cover Art (Back)"

    try:
        data = image.read()
    except OSError as e:
        raise AudioFileError(e) from e

    ext = (image.extensions and image.extensions[0]) or "jpg"
    data = f"hello.{ext}\x00".encode("ascii") + data

    value = mutagen.apev2.APEValue(data, mutagen.apev2.BINARY)

    return (key, value)


class APEv2File(AudioFile):
    # Map APE names to QL names. APE tags are also usually capitalized.
    # Also blacklist a number of tags.
    IGNORE = ["file", "index", "introplay", "dummy"]
    TRANS = {
        "subtitle": "version",
        "track": "tracknumber",
        "disc": "discnumber",
        "catalog": "labelid",
        "year": "date",
        "record location": "location",
        "album artist": "albumartist",
        "debut album": "originalalbum",
        "record date": "recordingdate",
        "original artist": "originalartist",
        "mixartist": "remixer",
    }
    SNART = {v: k for k, v in TRANS.items()}

    can_change_images = True

    def __init__(self, filename, audio=None):
        if audio:
            tag = audio.tags or {}
        else:
            with translate_errors():
                try:
                    tag = mutagen.apev2.APEv2(filename)
                except mutagen.apev2.APENoHeaderError:
                    tag = {}

        for key, value in tag.items():
            if get_cover_type(key, value) is not None:
                self.has_images = True

            key = self.TRANS.get(key.lower(), key.lower())
            if value.kind == mutagen.apev2.TEXT and key not in self.IGNORE:
                self[key] = "\n".join(list(value))

        self.sanitize(filename)

    @staticmethod
    def __titlecase(key):
        if key.lower() in ["isrc", "isbn", "ean/upc"]:
            return key.upper()
        return key.title()

    def can_change(self, key=None):
        if key is None:
            return True
        return (
            super().can_change(key)
            and key.lower() not in self.IGNORE
            and key.lower() not in self.TRANS
            and mutagen.apev2.is_valid_apev2_key(self.__titlecase(key))
        )

    def write(self):
        with translate_errors():
            try:
                tag = mutagen.apev2.APEv2(self["~filename"])
            except mutagen.apev2.APENoHeaderError:
                tag = mutagen.apev2.APEv2()

        # Remove any text keys we read in
        for key in list(tag.keys()):
            value = tag[key]
            if value.kind == mutagen.apev2.TEXT and key.lower() not in self.IGNORE:
                del tag[key]

        # Write all tags we have
        lower = self.as_lowercased()
        for key in lower.realkeys():
            new_key = self.SNART.get(key, key)
            if new_key in self.IGNORE:
                continue
            new_key = self.__titlecase(new_key)
            tag[new_key] = lower.list(key)

        with translate_errors():
            tag.save(self["~filename"])
        self.sanitize()

    def get_primary_image(self):
        try:
            tag = mutagen.apev2.APEv2(self["~filename"])
        except Exception:
            return None

        primary = None
        for key, value in tag.items():
            primary = (key, value)
            cover_type = get_cover_type(key, value)
            if cover_type == APICType.COVER_FRONT:
                break

        if primary is not None:
            return parse_cover(*primary)
        return None

    def get_images(self):
        try:
            tag = mutagen.apev2.APEv2(self["~filename"])
        except Exception:
            return []

        images = []
        for key, value in tag.items():
            image = parse_cover(key, value)
            if image is not None:
                images.append(image)

        images.sort(key=lambda c: c.sort_key)
        return images

    def clear_images(self):
        with translate_errors():
            try:
                tag = mutagen.apev2.APEv2(self["~filename"])
            except mutagen.apev2.APENoHeaderError:
                return

            for key, value in tag.items():
                cover_type = get_cover_type(key, value)
                if cover_type is not None:
                    del tag[key]

            tag.save()

        self.has_images = False

    def set_image(self, image):
        with translate_errors():
            try:
                tag = mutagen.apev2.APEv2(self["~filename"])
            except mutagen.apev2.APENoHeaderError:
                tag = mutagen.apev2.APEv2()

        for key, value in tag.items():
            cover_type = get_cover_type(key, value)
            if cover_type is not None:
                del tag[key]

        key, value = write_cover(image)
        tag[key] = value
        with translate_errors():
            tag.save(self["~filename"])

        self.has_images = True
